#
# Copyright 2019-2021 Lars Pastewka
#           2019-2021 Michael Röttger
#           2019-2020 Antoine Sanner
#           2019-2020 Kai Haase
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from collections import namedtuple

import dateutil.parser
import numpy as np

from ..Exceptions import MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography, UniformLineScan
from ..Support.UnitConversion import get_unit_conversion_factor, mangle_length_unit_utf8
from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo

MAGIC = b'VCA DATA\x01\x00\x00\x55'

DEKTAK_MATRIX = 0x00  # Too lazy to assign an actual type id?
DEKTAK_BOOLEAN = 0x01  # Takes value 0 and 1
DEKTAK_SINT32 = 0x06
DEKTAK_UINT32 = 0x07
DEKTAK_SINT64 = 0x0a
DEKTAK_UINT64 = 0x0b
DEKTAK_FLOAT = 0x0c  # Single precision float
DEKTAK_DOUBLE = 0x0d  # Double precision float
DEKTAK_TYPE_ID = 0x0e  # Compound type holding some kind of type id
DEKTAK_STRING = 0x12  # Free-form string value
DEKTAK_QUANTITY = 0x13  # Value with units (compound type)
DEKTAK_TIME_STAMP = 0x15  # Datetime (string/9-byte binary)
DEKTAK_UNITS = 0x18  # Units (compound type)
DEKTAK_DOUBLE_ARRAY = 0x40  # Raw data array, in XML Base64-encoded
DEKTAK_STRING_LIST = 0x42  # List of Str
DEKTAK_RAW_DATA = 0x46  # Parent/wrapper tag of raw data
DEKTAK_RAW_DATA_2D = 0x47  # Parent/wrapper tag of raw data
# Base64-encoded positions, not sure how it differs from 64
DEKTAK_POS_RAW_DATA = 0x7c
DEKTAK_CONTAINER = 0x7d  # General nested data structure
# Always the last item. Usually a couple of 0xff bytes inside.
DEKTAK_TERMINATOR = 0x7f

TIMESTAMP_SIZE = 9
UNIT_EXTRA = 12
DOUBLE_ARRAY_EXTRA = 5


class OPDxReader(ReaderBase):
    _format = 'opdx'
    _name = 'Dektak OPDx'
    _description = '''
File format of the Bruker Dektak XT* series stylus profilometer.
'''

    # Reads in the positions of all the data and metadata
    def __init__(self, file_path):
        self.file_path = file_path
        with OpenFromAny(file_path, 'rb') as f:
            # Check OPDx file magic
            if f.read(len(MAGIC)) != MAGIC:
                raise ValueError('File magic does not match. This is not a Dektak OPDx file.')

            # Read OPDx file manifest (without reading arrays and matrices)
            self.manifest = {}
            while _read_item(f, self.manifest):
                pass

        # Populate channel information
        self._channels = []
        data_kind = self.manifest['/MetaData/DataKind']
        if data_kind == 'Surface Profile':
            # This file contains a line scan
            self.read_linescan_channel_infos(0)
        elif data_kind == 'Surface Height':
            # This file contains a topography scan
            self.read_topography_channel_infos(0)
        else:
            raise ValueError(f"Don't know how to read data of kind '{data_kind}'.")

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={},
                   periodic=False, subdomain_locations=None,
                   nb_subdomain_grid_pts=None):
        if channel_index is None:
            channel_index = self._default_channel_index

        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError('This reader does not support MPI '
                               'parallelization.')

        channel_info = self._channels[channel_index]
        prefix = channel_info.info['opdx_prefix']
        if channel_info.unit is not None:
            if unit is not None:
                raise MetadataAlreadyFixedByFile('unit')
            unit = channel_info.unit
        if height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')

        info = info.copy()
        info.update(channel_info.info)

        if channel_info.dim == 1:
            position, length = self.manifest[f'{prefix}/Array']
            assert length == 8 * channel_info.nb_grid_pts[0] + DOUBLE_ARRAY_EXTRA
            with OpenFromAny(self.file_path, 'rb') as f:
                f.seek(position + DOUBLE_ARRAY_EXTRA)
                data = np.frombuffer(f.read(length - DOUBLE_ARRAY_EXTRA), np.dtype('<f8'))

            physical_sizes = self._check_physical_sizes(physical_sizes, channel_info.physical_sizes)

            return UniformLineScan(heights=data, physical_sizes=physical_sizes, unit=unit, periodic=periodic,
                                   info=info).scale(channel_info.height_scale_factor)
        elif channel_info.dim == 2:
            some_int, another_name, position, length, xres, yres = self.manifest[f'{prefix}/Matrix']
            assert channel_info.nb_grid_pts == (xres, yres)
            assert length == 4 * xres * yres

            with OpenFromAny(self.file_path, 'rb') as f:
                f.seek(position)
                data = np.frombuffer(f.read(length), np.dtype('<f4')).reshape((yres, xres)).T

            physical_sizes = self._check_physical_sizes(physical_sizes, channel_info.physical_sizes)

            return Topography(heights=data, physical_sizes=physical_sizes, unit=unit, periodic=periodic,
                              info=info).scale(channel_info.height_scale_factor)
        else:
            raise RuntimeError(f"Don't know how to read a {channel_info.dim}-dimensional dataset.")

    @property
    def channels(self):
        return self._channels

    def info_from_manifest(self, prefix):
        info = {'opdx_prefix': prefix}

        try:
            acquisition_time = str(dateutil.parser.parse(
                self.manifest['/MetaData/Date'] + ' ' + self.manifest['/MetaData/Time']))
            info['acquisition_time'] = acquisition_time
        except KeyError:
            pass

        try:
            info['instrument'] = {'name': self.manifest['/MetaData/MeasurementSettings/InstrumentName']}
        except KeyError:
            pass

        return info

    def read_linescan_channel_infos(self, channel_index):
        """
        Read line scan (profile) information
        """
        channel_names = self.manifest['/MetaData/1D_Channels/Height']

        for channel_name in channel_names:
            prefix = f'/1D_Data/{channel_name}'

            nb_grid_pts = self.manifest[f'{prefix}/NumPoints']
            physical_size = self.manifest[f'{prefix}/Extent'].value
            unit = self.manifest[f'{prefix}/Extent'].symbol

            height_scale_factor = self.manifest[f'{prefix}/DataScale'].value
            height_unit = self.manifest[f'{prefix}/DataScale'].symbol

            height_scale_factor *= get_unit_conversion_factor(height_unit, unit)

            self._channels += [ChannelInfo(self, channel_index,
                                           name=self.manifest[f'{prefix}/DataKind'],
                                           dim=1,
                                           nb_grid_pts=nb_grid_pts,
                                           physical_sizes=physical_size,
                                           uniform=True,
                                           unit=unit,
                                           height_scale_factor=height_scale_factor,
                                           info=self.info_from_manifest(prefix))]

            channel_index += 1

    def read_topography_channel_infos(self, channel_index):
        """
        Read topography (2D map) information
        """
        channel_name = self.manifest['/MetaData/PrimaryData2D']
        prefix = f'/2D_Data/{channel_name}'

        nb_grid_pts_y = self.manifest[f'{prefix}/Dimension1Points']
        nb_grid_pts_x = self.manifest[f'{prefix}/Dimension2Points']

        physical_size_y = self.manifest[f'{prefix}/Dimension1Extent'].value
        unit_y = self.manifest[f'{prefix}/Dimension1Extent'].symbol

        physical_size_x = self.manifest[f'{prefix}/Dimension2Extent'].value
        unit_x = self.manifest[f'{prefix}/Dimension2Extent'].symbol

        physical_size_y *= get_unit_conversion_factor(unit_y, unit_x)

        height_scale_factor = self.manifest[f'{prefix}/DataScale'].value
        height_unit = self.manifest[f'{prefix}/DataScale'].symbol

        height_scale_factor *= get_unit_conversion_factor(height_unit, unit_x)

        self._channels += [ChannelInfo(self, channel_index,
                                       name=self.manifest[f'{prefix}/DataKind'],
                                       dim=2,
                                       nb_grid_pts=(nb_grid_pts_x, nb_grid_pts_y),
                                       physical_sizes=(physical_size_x, physical_size_y),
                                       uniform=True,
                                       unit=unit_x,
                                       height_scale_factor=height_scale_factor,
                                       info=self.info_from_manifest(prefix))]

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__


DektakUnit = namedtuple('DektakUnit', ['name', 'symbol', 'value', 'extra'])


def _read_item(stream, manifest, prefix='', offset=0):
    """
    Reads in the next item out of the buffer and saves it in the manifest.
    May recursively call itself for containers.

    Parameters
    ----------
    stream : file-like stream object
        The input file/buffer.
    manifest : dict
        The manifest that will be updated by a call to this function.
    prefix : str, optional
        Prefix to the keys that are stored in the manifest. (Default: '')
    offset : int, optional
        Offset of f relative to the absolute start of the buffer holding the
        full OPDx.
        (Default: 0)

    Returns
    -------
    key : str
        Return the key for the item that was created by this read command.
        None indicates that the end of the buffer or file has been reached.
    """
    name = _read_name(stream)
    path = f'{prefix}/{name}'
    data = None

    typeid = _read_scalar(stream, 'B')

    # Container types
    if typeid == DEKTAK_CONTAINER or typeid == DEKTAK_RAW_DATA or typeid == DEKTAK_RAW_DATA_2D:
        length = _read_varlen(stream)
        start = stream.tell()
        while _read_item(stream, manifest, prefix=path):
            pass
        # There are two bytes at the end of every container with unknown purpose...
        _read_scalar(stream, '<u2')
        if start + length != stream.tell():
            raise RuntimeError(f'The OPDx reader has missed some data entries. The stream position {stream.tell()} '
                               f'does not equal the expected position {start + length}.')
        stream.seek(start + length)
        return path
    elif typeid == DEKTAK_POS_RAW_DATA:
        if path.startswith('/2D_Data'):
            _read_name(stream)  # typename, we discard this information
            length = _read_varlen(stream)
            start = stream.tell()
            unitx, divisorx = read_dimension2d_content(stream)
            unity, divisory = read_dimension2d_content(stream)
            data = (unitx, divisorx, unity, divisory)
            if start + length != stream.tell():
                raise RuntimeError(f'The OPDx reader has missed some data entries. The stream position {stream.tell()} '
                                   f'does not equal the expected position {start + length}.')
        elif path.startswith('/1D_Data'):
            _read_name(stream)  # typename, we discard this information
            length = _read_varlen(stream)
            start = stream.tell()
            unit = _read_unit_data(stream)
            count = _read_scalar(stream, '<u8')
            # Skip over data
            stream.seek(start + length)
            data = (unit, count)
        else:
            # TODO check if should assume 1D here like Gwyddion
            raise ValueError
    elif typeid == DEKTAK_TERMINATOR:
        # Terminator
        return None
    else:
        try:
            t = _item_readers[typeid]
        except KeyError:
            raise ValueError(f"Don't know how to read type with id {typeid}.")
        if isinstance(t, str):
            data = _read_scalar(stream, t)
        else:
            data = t(stream)

    if data is not None:
        manifest[path] = data
        return path
    else:
        return None


def _read_time_stamp(stream):
    return stream.read(TIMESTAMP_SIZE)


def _read_unit(stream):
    """
    Reads in a quantity unit: Value, name and symbol.

    Parameters
    ----------
    f : bytes
        The input buffer

    Returns
    -------
    unit : DektakUnit
        A unit item, filled with value, name and symbol
    """
    length = _read_varlen(stream)
    start = stream.tell()
    unit = _read_unit_data(stream)
    if start + length != stream.tell():
        raise RuntimeError(f'The OPDx reader has missed some data entries. The stream position {stream.tell()} '
                           f'does not equal the expected position {start + length}.')
    return unit


def _read_unit_data(stream):
    """
    Reads in a quantity unit: Value, name and symbol.

    Parameters
    ----------
    stream : bytes
        The input buffer

    Returns
    -------
    unit : DektakUnit
        A unit item, filled with value, name and symbol
    """
    name = _read_name(stream)
    symbol = mangle_length_unit_utf8(_read_name(stream))

    value = _read_scalar(stream, '<f8')
    extra = stream.read(UNIT_EXTRA)

    return DektakUnit(name, symbol, value, extra)


def _read_quantity(stream):
    """
    Reads in a quantity unit: Value, name and symbol.

    Parameters
    ----------
    stream : bytes
        The input buffer

    Returns
    -------
    unit : DektakUnit
        A unit item, filled with value, name and symbol
    """
    length = _read_varlen(stream)
    start = stream.tell()

    value = _read_scalar(stream, '<f8')
    name = _read_name(stream)
    symbol = mangle_length_unit_utf8(_read_name(stream))

    stream.seek(start + length)

    return DektakUnit(name, symbol, value, [])


def read_dimension2d_content(stream):
    """
    Reads in information about a 2d dimension.

    Parameters
    ----------
    stream : bytes
        The input buffer

    Returns
    -------
    unit : DektakUnit
        The unit
    divisor : float
        Divisor
    """
    value = _read_scalar(stream, '<f8')
    name = _read_name(stream)
    symbol = _read_name(stream)
    divisor = _read_scalar(stream, '<f8')
    extra = stream.read(UNIT_EXTRA)
    return DektakUnit(name, symbol, value, extra), divisor


def _read_name(stream):
    # Names always have a size of 4 bytes
    length = _read_scalar(stream, '<u4')
    return stream.read(length).decode('UTF-8')


def _read_string(stream):
    # String have variable lengths
    string_length = _read_varlen(stream)
    return stream.read(string_length).decode('UTF-8')


def _read_scalar(stream, dtype):
    dtype = np.dtype(dtype)
    buffer = stream.read(dtype.itemsize)
    if buffer:
        return np.frombuffer(buffer, dtype=dtype, count=1)[0]
    else:
        return None


def _read_varlen(stream):
    lenlen = _read_scalar(stream, 'B')
    if lenlen == 1:
        return _read_scalar(stream, 'B')
    elif lenlen == 2:
        return _read_scalar(stream, '<u2')
    elif lenlen == 4:
        return _read_scalar(stream, '<u4')
    else:
        raise ValueError(f"Don't know how to read a variable length of size {lenlen}.")


def _read_string_list(stream):
    _read_name(stream)  # typename, we discard this information
    length = _read_varlen(stream)
    start = stream.tell()
    data = []
    while stream.tell() < start + length:
        s = _read_name(stream)
        data += [s]
    if start + length != stream.tell():
        raise RuntimeError(f'The OPDx reader has missed some data entries. The stream position {stream.tell()} '
                           f'does not equal the expected position {start + length}.')
    return data


def _read_double_array(stream):
    _read_name(stream)  # typename, we discard this information
    length = _read_varlen(stream)
    data = (stream.tell(), length)
    # Skip over data
    stream.seek(length, 1)
    return data  # This is start position and length of the buffer


def _read_type_id(stream):
    _read_name(stream)  # typename, we discard this information
    length = _read_varlen(stream)
    return stream.read(length)


def _read_matrix(stream):
    _read_name(stream)  # typename, we discard this information
    some_int = _read_scalar(stream, '<u4')
    another_name = _read_name(stream)
    length = _read_varlen(stream)
    yres = _read_scalar(stream, '<u4')
    xres = _read_scalar(stream, '<u4')
    if length < 8:  # 2 * sizeof int32
        raise ValueError
    length -= 8  # Remove xres and yres
    data = (some_int, another_name, stream.tell(), length, xres, yres)
    # Skip over data
    stream.seek(length, 1)
    return data


_item_readers = {
    DEKTAK_MATRIX: _read_matrix,
    DEKTAK_BOOLEAN: '?',
    DEKTAK_SINT32: '<i4',
    DEKTAK_UINT32: '<u4',
    DEKTAK_SINT64: '<i8',
    DEKTAK_UINT64: '<u8',
    DEKTAK_FLOAT: '<f4',
    DEKTAK_DOUBLE: '<f8',
    DEKTAK_TYPE_ID: _read_type_id,
    DEKTAK_STRING: _read_string,
    DEKTAK_QUANTITY: _read_quantity,
    DEKTAK_TIME_STAMP: _read_time_stamp,
    DEKTAK_UNITS: _read_unit,
    DEKTAK_DOUBLE_ARRAY: _read_double_array,
    DEKTAK_STRING_LIST: _read_string_list,
}



#
# Copyright 2019-2021 Lars Pastewka
#           2019-2021 Michael Röttger
#           2019-2021 Antoine Sanner
#           2019 Kai Haase
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import abc

import numpy as np

from ..Exceptions import MetadataAlreadyFixedByFile


class ChannelInfo:
    """
    Information on topography channels contained within a file.
    """

    def __init__(self, reader, index, name=None, dim=None, nb_grid_pts=None, physical_sizes=None,
                 height_scale_factor=None, periodic=None, uniform=None, undefined_data=None, unit=None, info={}):
        """
        Initialize the channel. Use as many information from the file as
        possible by passing it in the keyword arguments. Keyword arguments
        can be None if the information cannot be determined. (This is the
        default for all keyword arguments.)

        Arguments
        ---------
        reader : :obj:`ReaderBase`
            Reader instance this channel is coming from.
        index : int
            Index of channel in the file, where zero is the first channel.
        name : str, optional
            Name of the channel. If no name is given, "channel <index>" will
            be used, where "<index>" is replaced with the index.
        dim : int, optional
            Number of dimensions.
        nb_grid_pts : tuple of ints, optional
            Number grid points in each dimension.
        physical_sizes : tuple of floats, optional
            Physical dimensions.
        height_scale_factor: float, optional
            Number by which all heights have been multiplied.
        periodic : bool, optional
            Whether the SurfaceTopography should be interpreted as one period of
            a periodic surface. This will affect the PSD and autocorrelation
            calculations (windowing).
        uniform : bool, optional
            Data is uniform.
        has_undefined_data : bool, optional
            Underlying data has missing/undefined points.
        unit : str, optional
            Length unit of measurement.
        info : dict, optional
            Meta data found in the file. (Default: {})
        """
        self._reader = reader
        self._index = int(index)
        self._name = "channel {}".format(self._index) \
            if name is None else str(name)
        self._dim = dim
        self._nb_grid_pts = None \
            if nb_grid_pts is None else tuple(np.ravel(nb_grid_pts))
        self._physical_sizes = None \
            if physical_sizes is None else tuple(np.ravel(physical_sizes))
        self._height_scale_factor = height_scale_factor
        self._periodic = periodic
        self._uniform = uniform
        self._undefined_data = undefined_data
        self._unit = unit
        if info is None:
            self._info = None
        else:
            self._info = info.copy()

    def topography(self, physical_sizes=None, height_scale_factor=None, unit=None, info={}, periodic=False,
                   subdomain_locations=None, nb_subdomain_grid_pts=None):
        """
        Returns an instance of a subclass of :obj:`HeightContainer` that
        contains the topography data. The method allows to override data
        found in the data file.

        Arguments
        ---------
        physical_sizes : tuple of floats, optional
            Physical size of the topography. It is necessary to specify this
            if no physical size is found in the data file. If there is a
            physical size, then this parameter will override the physical
            size found in the data file.
        height_scale_factor : float, optional
            Factor by which the heights should be multiplied.
            This parameter is only available for file formats that do not provide
            metadata information about units and associated length scales.
        unit : str, optional
            Length unit of measurement.
        info : dict, optional
            This dictionary will be appended to the info dictionary returned
            by the reader.
        periodic : bool, optional
            Whether the SurfaceTopography should be interpreted as one period of
            a periodic surface. This will affect the PSD and autocorrelation
            calculations (windowing)
        subdomain_locations : tuple of ints, optional
            Origin (location) of the subdomain handled by the present MPI
            process.
        nb_subdomain_grid_pts : tuple of ints, optional
            Number of grid points within the subdomain handled by the present
            MPI process.

        Returns
        -------
        topography : subclass of :obj:`HeightContainer`
            The object containing the actual topography data.
        """
        return self._reader.topography(
            channel_index=self._index,
            physical_sizes=physical_sizes,
            height_scale_factor=height_scale_factor,
            unit=unit,
            info=info,
            periodic=periodic,
            subdomain_locations=subdomain_locations,
            nb_subdomain_grid_pts=nb_subdomain_grid_pts)

    @property
    def index(self):
        """Unique integer channel index."""
        return self._index

    @property
    def name(self):
        """
        Name of the channel.
        Can be used in a UI for identifying a channel.
        """
        return self._name

    @property
    def dim(self):
        """
        1 for line scans and 2 for topography maps.
        """
        return self._dim

    @property
    def nb_grid_pts(self):
        """
        Number of grid points in each direction, either 1 or 2 elements
        depending on the dimension of the topography.
        """
        return self._nb_grid_pts

    @property
    def physical_sizes(self):
        """
        If the physical size can be determined from the file, a tuple is
        returned. The tuple has 1 or 2 elements (size_x, size_y), depending
        on the dimension of the topography.

        If the physical size can not be determined from the file, then
        None is returned.

        Note that the physical sizes obtained from the file can be
        overwritten by passing a `physical_sizes` argument to the
        `topography` method that returns the topography object.
        """
        return self._physical_sizes

    @physical_sizes.setter
    def physical_sizes(self, value):
        self._physical_sizes = value

    @property
    def height_scale_factor(self):
        """
        If a height scale factor can be determined from the file, a float is
        returned.

        If no height scale factor can be determined from the file, then
        None is returned.

        Note that the height scale factor obtained from the file can be
        overwritten by passing a `height_scale_factor` argument to the
        `topography` method that returns the topography object.
        """
        return self._height_scale_factor

    @height_scale_factor.setter
    def height_scale_factor(self, value):
        self._height_scale_factor = value

    @property
    def is_periodic(self):
        """
        Return whether the topography is periodically repeated at the
        boundaries.
        """
        return self._periodic

    @property
    def is_uniform(self):
        """
        Return whether the topography is uniform.
        """
        return self._uniform

    @property
    def has_undefined_data(self):
        """
        Return whether the topography has undefined data.
        """
        return self._undefined_data

    @property
    def pixel_size(self):
        """
        The pixel size is returned as the phyiscal size divided by the number
        of grid points. If the physical size can be determined from the file,
        a tuple is returned. The tuple has 1 or 2 elements (size_x, size_y),
        depending on the dimension of the topography.

        If the physical size can not be determined from the file, then
        None is returned.

        Note that the physical sizes obtained from the file can be overwritten
        by passing a `physical_sizes` argument to the `topography` method that
        returns the topography object. Overriding the physical size will also
        affect the pixel size.
        """
        if self._physical_sizes is None or self._nb_grid_pts is None:
            return None
        else:
            return tuple(np.array(self._physical_sizes) /
                         np.array(self._nb_grid_pts))

    @property
    def area_per_pt(self):
        """
        The area per point is returned as the product over the pixel size
        tuple. If `pixel_size` returns None, than also `area_per_pt` returns
        None.
        """
        if self.pixel_size is None:
            return None
        else:
            return np.prod(self.pixel_size)

    @property
    def unit(self):
        """
        Length unit.
        """
        return self._unit

    @property
    def info(self):
        """
        A dictionary containing additional information (metadata) not used by
        SurfaceTopography itself, but required by third-party application.

        Presently, the following entries have been standardized:
        'unit':
            This is the unit of the physical size (if given in the file) and
            the units of the heights.
        'height_scale_factor':
            Factor which was used to scale the raw numbers from file (which
            can be voltages or some other quantity actually acquired in the
            measurement technique) to heights with the given 'unit'.
        """
        info = self._info.copy()
        if self.unit is not None:
            info.update(dict(unit=self.unit))
        return info


class ReaderBase(metaclass=abc.ABCMeta):
    """
    Base class for topography readers. These are object that allow to open a
    file (from filename or stream object), inspect its metadata and then
    request to load a topography from it. Metadata is typically extracted
    without reading the full file.

    Readers should adhere to the following design rules:
    1. Opening a file should be fast and therefore not read the whole data.
       The data is read only when requesting it via the `topography` method.
    2. Data corruption must be detected when opening the file. The
       `topography` method must not fail because of file corruption issues.
    These rules are important to allow smooth operation of the readers in
    the web application `TopoBank`.
    """

    _format = None
    _name = None
    _description = None

    _default_channel_index = 0

    @classmethod
    def format(cls):
        """
        String identifier for this file format. Identifier must be unique and
        is typically equal to the file extension of this format.
        """
        if cls._format is None:
            raise RuntimeError('Reader does not provide a format string')
        return cls._format

    @classmethod
    def name(cls):
        """
        Short name of this file format.
        """
        if cls._name is None:
            raise RuntimeError('Reader does not provide a name')
        return cls._name

    @classmethod
    def description(cls):
        """
        Long description of this file format. Should be formatted as markdown.
        """
        if cls._description is None:
            raise RuntimeError('Reader does not provide a description string')
        return cls._description

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        pass

    @property
    @abc.abstractmethod
    def channels(self):
        """
        Returns a list of :obj:`ChannelInfo`s describing the available data
        channels.
        """
        raise NotImplementedError

    @property
    def default_channel(self):
        """Return the default channel. This is often the first channel with
        height information."""
        return self.channels[self._default_channel_index]

    @classmethod
    def _check_physical_sizes(cls, physical_sizes_from_arg,
                              physical_sizes=None):
        """Handle overriding of `physical_sizes` arguments and make sure,
        that return value is not `None`.

        Parameters
        ----------
        physical_sizes_from_arg: tuple or float or None
            physical sizes given as argument when instantiating a topography from a reader
        physical_sizes: tuple or float or None
            physical sizes which were present in the file
        """
        if physical_sizes is None:
            if physical_sizes_from_arg is None:
                raise ValueError("`physical_sizes` could not be extracted from file, you must provide it.")
            eff_physical_sizes = physical_sizes_from_arg
        elif physical_sizes_from_arg is not None:
            # in general this should result in an exception now
            raise MetadataAlreadyFixedByFile('physical_sizes')
        else:
            eff_physical_sizes = physical_sizes

        return eff_physical_sizes

    @abc.abstractmethod
    def topography(self, channel_index=None, physical_sizes=None, height_scale_factor=None, unit=None, info={},
                   periodic=False, subdomain_locations=None, nb_subdomain_grid_pts=None):
        """
        Returns an instance of a subclass of :obj:`HeightContainer` that
        contains the topography data. Which specific type of height container
        (1D, 2D, uniform, nonuniform) is returned may depend on the file
        content and is determined dynamically.

        The returned object needs to have `physical_sizes` and periodicity
        information. If this information is not provided by the data file,
        then the user must specify it when calling this method. Conversely,
        metadata that is present in the file cannot be overridden but needs to
        lead to a `MetadataAlreadyFixedByFile` exception. The user can check
        if the metadata is present by inspecting the `ChannelInfo` of the
        respective data channel.

        Arguments
        ---------
        channel_index : int
            Index of the channel to load. See also `channels` method.
            (Default: None, which load the default channel)
        physical_sizes : tuple of floats
            Physical size of the topography. It is necessary to specify this
            if no physical size is found in the data file. If there is a
            physical size, then this parameter will override the physical
            size found in the data file.
        height_scale_factor : float
            Override height scale factor found in the data file.
        unit : str
            Length unit.
        info : dict
            This dictionary will be appended to the info dictionary returned
            by the reader.
        periodic: bool
            Whether the SurfaceTopography should be interpreted as one period of
            a periodic surface. This will affect the PSD and autocorrelation
            calculations (windowing).
        subdomain_locations : tuple of ints
            Origin (location) of the subdomain handled by the present MPI
            process.
        nb_subdomain_grid_pts : tuple of ints
            Number of grid points within the subdomain handled by the present
            MPI process.

        Returns
        -------
        topography : subclass of :obj:`HeightContainer`
            The object containing the actual topography data.

        Raises
        ------
        MetadataAlreadyFixedByFile
            Raised if `physical_sizes`, `unit or `height_scale_factor` have
            already been defined in the file, because they should not be
            overridden by the user.
        """
        raise NotImplementedError



#
# Copyright 2019-2021 Lars Pastewka
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from ..UniformLineScanAndTopography import Topography
from .Reader import ReaderBase, ChannelInfo


class H5Reader(ReaderBase):
    _format = 'h5'
    _name = 'Hierarchical data format (HDF5)'
    _description = '''
Import filter for [HDF5](https://support.hdfgroup.org/HDF5/) files provided
within the contact mechanics challenge. The reader looks for a two-dimensional
array named `surface`. HDF5 files do not store units or physical sizes. These
need to be manually provided by the user.

The original contact mechanics challenge data can be downloaded
[here](https://www.lmp.uni-saarland.de/index.php/research-topics/contact-mechanics-challenge-announcement/).
    '''  # noqa: E501

    def __init__(self, fobj):
        self._h5 = None
        import h5py
        self._h5 = h5py.File(fobj, 'r')
        self._channels = []
        channel_index = 0
        for name in self._h5:
            if len(self._h5[name].shape) == 2:
                # This looks like a topography
                self._channels += [ChannelInfo(self,
                                               channel_index,  # channel index
                                               name=name,
                                               dim=len(self._h5[name].shape),
                                               uniform=True,
                                               nb_grid_pts=self._h5[name].shape)]
                channel_index += 1

    def close(self):
        if self._h5 is not None:
            self._h5.close()

    @property
    def channels(self):
        return self._channels

    def topography(self, channel_index=None, physical_sizes=None, height_scale_factor=None, unit=None, info={},
                   periodic=False, subdomain_locations=None, nb_subdomain_grid_pts=None):
        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError(
                'This reader does not support MPI parallelization.')
        if channel_index is None:
            channel_index = self._default_channel_index
        channel = self._channels[channel_index]
        size = self._check_physical_sizes(physical_sizes)
        t = Topography(self._h5[channel.name][...], size, unit=unit, info=info, periodic=periodic)
        if height_scale_factor is not None:
            t = t.scale(height_scale_factor)
        return t

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__



#
# Copyright 2016-2021 Lars Pastewka
#           2020 Michael Röttger
#           2018-2020 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import glob
from pathlib import Path
import re

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext

class CustomBuildExtCommand(build_ext):
    """build_ext command for use when numpy headers are needed."""

    def run(self):
        # Import numpy here, only when headers are needed
        import numpy

        # Add numpy headers to include_dirs
        self.include_dirs.append(numpy.get_include())

        # Call original build_ext command
        build_ext.run(self)


#
# Optain LAPACK configuration from numpy
#

try:
    import numpy as np
except ImportError:
    print("WARNING: Could not find numpy, skipping LAPACK detection.")
    np = None

lib_dirs = []
libs = []
extra_link_args = []
extra_objects = []
if np is not None:
    for k, v in np.__config__.__dict__.items():
        if re.match('lapack_.*_info', k):
            if v:
                print("* Using LAPACK information from '{}' dictionary in " \
                    "numpy.__config__".format(k))
                try:
                    print("    library_dirs = {}".format(v['library_dirs']))
                    lib_dirs += v['library_dirs']
                except:
                    print("    No 'library_dirs' entry found.")
                try:
                    print("    libraries = {}".format(v['libraries']))
                    libs += v['libraries']
                except:
                    print("    No 'libraries' entry found.")
                try:
                    print("    extra_link_args = '{}'".format(v['extra_link_args']))
                    extra_link_args += v['extra_link_args']
                except:
                    print("    No 'extra_link_args' entry found.")
                # We use whichever lapack_*_info comes first, hopefully there is
                # just one.
                break
    if len(libs) == 0:
        # No entries from np.__config__
        print("* Using lapack_lite")
        extra_objects += [glob.glob(np.linalg.__path__[0]+'/lapack_lite*.so')[0]]


extra_compile_args = ["-std=c++11"]
print(extra_objects)

scripts = []

extensions = [
    Extension(
        name='_SurfaceTopography',
        sources=['c/autocorrelation.cpp',
                 'c/bicubic.cpp',
                 'c/patchfinder.cpp',
                 'c/module.cpp'],
        extra_compile_args=extra_compile_args,
        library_dirs=lib_dirs,
        libraries=libs,
        extra_link_args=extra_link_args,
        extra_objects=extra_objects
    )
]

setup(
    name="SurfaceTopography",
    cmdclass={'build_ext': CustomBuildExtCommand} if np else {},
    scripts=scripts,
    packages=find_packages(),
    package_data={'': ['ChangeLog.md']},
    include_package_data=True,
    ext_modules=extensions,
    # metadata for upload to PyPI
    author='Lars Pastewka',
    author_email='lars.pastewka@imtek.uni-freiburg.de',
    description='Read and analyze surface topographies',
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type='text/markdown',
    license='MIT',
    test_suite='tests',
    # dependencies
    python_requires='>=3.5.0',
    use_scm_version=True,
    zip_safe=False,
    setup_requires=[
        'setuptools_scm>=3.5.0',
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
    ],
    install_requires=[
        'numpy>=1.16.3',
        'NuMPI>=0.3.0',
        'muFFT>=0.12.0',
        'igor',
        'h5py',
        'defusedxml',
        'numpyencoder',
        'openpyxl',
        'pyyaml',
        'Pillow',
        'requests',
        'matplotlib>=1.0.0',
        'python-dateutil',
        'tiffile',
    ]
)



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Compute properties as averages over individual results."""

import logging

import numpy as np
from collections import defaultdict

from ..Exceptions import NoReliableDataError, UndefinedDataError
from ..Support.Regression import resample
from .SurfaceContainer import SurfaceContainer

_log = logging.Logger(__name__)


def log_average(self, function_name, unit, nb_points_per_decade=10, reliable=True, progress_callback=None, **kwargs):
    """
    Return average property on a logarithmic grid.

    Parameters
    ----------
    self : SurfaceContainer
        Container object containing the underlying data sets.
    function_name : str
        Function for which to compute the average, e.g.
        'autocorrelation_from_profile'.
    unit : str
        Unit of the distance array. All topographies are converted to this
        unit before the properties are computed.
    nb_points_per_decade : int, optional
        Number of points per decade in length for automatic grid construction.
        (Default: 10)
    reliable : bool, optional
        Only return data deemed reliable. (Default: True)
    progress_callback : func, optional
        Function taking iteration and the total number of iterations as
        arguments for progress reporting. (Default: None)
    """
    results = defaultdict(list)
    for i, topography in enumerate(self):
        if progress_callback is not None:
            progress_callback(i, len(self))

        topography = topography.to_unit(unit)

        # Compute property... but do not allow any resampling
        func = getattr(topography, function_name)
        try:
            x, y = func(reliable=reliable, resampling_method=None, **kwargs)
            has_data = True
        except NoReliableDataError:
            has_data = False
        except UndefinedDataError:
            has_data = False

        # Construct grid in this range
        if has_data:
            # Idiot check: Is there actually more than one data point
            # (if not, NoReliableDataError should actually have been raised above)
            m = x > 0
            has_data = np.sum(m) >= 1
        if not has_data:
            _log.warning(f'Topography {topography} contributes no data to average.')
            continue
        lower, upper = np.min(x[m]), np.max(x)
        lower_decade, upper_decade = int(np.floor(np.log10(lower))), int(np.ceil(np.log10(upper)))

        # Resample onto the same grid
        collocation_points, bin_edges, resampled_y, variance = resample(
            x, y, collocation='log', min_value=10 ** lower_decade, max_value=10 ** upper_decade,
            nb_points=(upper_decade - lower_decade) * nb_points_per_decade + 1)

        # Store for later averaging
        for j, (xval, yval) in enumerate(zip(collocation_points, resampled_y)):
            results[lower_decade * nb_points_per_decade + j] += [(xval, yval)]

    if progress_callback is not None:
        progress_callback(len(self), len(self))

    if len(results) == 0:
        raise NoReliableDataError('Container contains no reliable data.')

    avgresults = np.array(sorted([np.nanmean(vals, axis=0) for vals in results.values()], key=lambda x: x[0]))
    return avgresults.T[0], avgresults.T[1]


# All container functionsa are 'profile' functions, because container can contain mixtures of topgoraphic maps and line
# scans.
SurfaceContainer.register_function(
    'autocorrelation', lambda self, unit, **kwargs: log_average(self, 'autocorrelation_from_profile', unit, **kwargs))
SurfaceContainer.register_function(
    'power_spectrum', lambda self, unit, **kwargs: log_average(self, 'power_spectrum_from_profile', unit, **kwargs))
SurfaceContainer.register_function(
    'variable_bandwidth',
    lambda self, unit, **kwargs: log_average(self, 'variable_bandwidth_from_profile', unit, **kwargs))



#
# Copyright 2018-2020 Lars Pastewka
#           2019-2020 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Tests to understand the difficulties in extracting hurst from noisy data
"""
import numpy as np
import scipy
import matplotlib.pyplot as plt

import PyCo.Tools as Tools
import SurfaceTopography as Surf


def plot_naive(surface, lam_max):
    fig = plt.figure()
    ax=fig.add_subplot(111)
    ax.set_yscale('log')
    ax.set_xscale('log')

    surf = Tools.CharacterisePeriodicSurface(surface)
    q = surf.q
    C = surf.C
    H, alpha = surf.estimate_hurst_naive(lambda_max=lam_max, full_output=True)
    print("H = {}, alpha = {}".format(H, alpha))
    ax.loglog(q, C, alpha=.1)
    mean, err, q_g = surf.grouped_stats(100)
    mask = np.isfinite(mean)
    mean = mean[mask]
    err = err[:, mask]
    q_g = q_g[mask]

    ax.errorbar(q_g, mean, yerr=err)
    ax.set_title("Naive: H={:.2f}, h_rms={:.2e}".format(H, np.sqrt((surface.heights() ** 2).mean())))
    a, b = np.polyfit(np.log(q), np.log(C), 1)
    ax.plot(q, q**(-2-2*H)*alpha, label="{}, H={:.2f}".format('fit', H))
    ax.legend(loc='best')

def plot_grad_C0(surface, H_in, lam_max):
    surf = Tools.CharacterisePeriodicSurface(surface)
    q_min = 2*np.pi/lam_max
    sl = surf.q > q_min
    q = surf.q[sl]
    C = surf.C[sl]
    dim = 2

    def C0_of_H(H):
        return ((q**(-3-2*H)).sum() /
            (q**(-5-4*H)/C).sum())

    def objective(H, C0):
        return ((1 - C0*q**(-2*H-2)/C)**2 /
                q**(dim-1)).sum()

    C0 = C0_of_H(H_in)
    O0 = objective(H_in, C0)
    c_s = np.linspace(0, 2*C0, 51)
    o_s = np.zeros_like(c_s)

    for i, c in enumerate(c_s):
        o_s[i] = objective(H_in, c)

    fig = plt.figure()
    ax=fig.add_subplot(111)
    fig.suptitle('grad(C0)')
    ax.plot(c_s, o_s, marker= '+')
    ax.scatter(C0, O0, marker='x', label = 'root', c='r')
    ax.grid(True)
    print("C0 = {}, obj0 = {}".format(C0, O0))
    return C0

def plot_grad_H(surface, lam_max):
    surf = Tools.CharacterisePeriodicSurface(surface)
    q_min = 2*np.pi/lam_max
    sl = surf.q > q_min
    q = surf.q[sl]# np.array(surf.q[sl][0], surf.q[sl][-1])
    C = surf.C[sl]# np.array(surf.C[sl][0], surf.C[sl][-1])
    dim = 2

    def C0_of_H(H):
        return ((C**2/q**(-5-dim-4*H)).sum() /
            (C/q**(-3-dim-2*H)).sum())

    def grad_h(H, C0):
        return (4*C0/C*np.log(q)*q**(-1-2*H-dim)*(1 - C0*q**(-2-2*H)/C)).sum()

    def objective(H, C0):
        return ((c/q**(-2*H-2) - C0)**2 /
                q**(dim-1)).sum()

    def full_obj(H):
        C0 = C0_of_H(H)
        return ((1 - C0/C*q**(-2*H-2))**2 /
                q**(dim-1)).sum()

    h_s = np.linspace(.0, 2., 51)
    o_s = np.zeros_like(h_s)
    g_s = np.zeros_like(h_s)

    for i, h in enumerate(h_s):
        c = C0_of_H(h)
        o_s[i] = objective(h, c)
        g_s[i] = grad_h(h, c)

    H_opt, obj_opt, err, nfeq = scipy.optimize.fminbound(full_obj, 0, 2, full_output=True)
    if err != 0:
        raise Exception()

    fig = plt.figure()
    ax=fig.add_subplot(211)
    ax.set_xlim(h_s[0], h_s[-1])
    fig.suptitle('grad(H)')
    ax.plot(h_s, o_s, marker= '+')
    ax.grid(True)
    ax.scatter(H_opt, obj_opt, marker='x', label = 'root', c='r')
    ax=fig.add_subplot(212)
    ax.set_xlim(h_s[0], h_s[-1])

    ax.plot(h_s, g_s, marker= '+')
    grad_opt = grad_h(H_opt, C0_of_H(H_opt))
    ax.scatter(H_opt, grad_opt, marker='x', label = 'root', c='r')
    #res = scipy.optimize.fmin
    #print("H_out = {}, obj0 = {}".format(C0, O0))
    ax.grid(True)

    return H_opt, C0_of_H(H_opt)

def compare_to_PyPy(surface, lam_max, H_ref, C0_ref):
    fig = plt.figure()
    ax=fig.add_subplot(111)
    ax.set_yscale('log')
    ax.set_xscale('log')

    surf = Tools.CharacterisePeriodicSurface(surface)
    q_min = 2*np.pi/lam_max
    sl = surf.q > q_min
    q = surf.q
    C = surf.C
    H, alpha, res = surf.estimate_hurst_alt(lambda_max=lam_max, full_output=True)
    print("H = {}, alpha = {}".format(H, alpha))
    ax.loglog(q, C, alpha=.1)
    mean, err, q_g = surf.grouped_stats(100)
    mask = np.isfinite(mean)
    mean = mean[mask]
    err = err[:, mask]
    q_g = q_g[mask]

    ax.errorbar(q_g, mean, yerr=err)
    ax.set_title("New: H_pypy={:.2f}, H_ref = {:.2f}, h_rms={:.2e}".format(H, H_ref, np.sqrt((surface.heights() ** 2).mean())))
    
    ax.plot(q[sl], q[sl]**(-2-2*H)*alpha, label="{}, H={:.4f}".format('fit', H), lw = 3)
    ax.plot(q[sl], q[sl]**(-2-2*H_ref)*C0_ref, label="{}, H={:.4f}".format('ref_fit', H_ref), lw = 3)
    ax.legend(loc='best')

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.loglog(q[sl], C[sl]/(q[sl]**(-2-2*H_ref)*C0_ref), alpha=.1)
    ax.errorbar(q_g, mean/(q_g**(-2-2*H_ref)*C0_ref), yerr=err/(q_g**(-2-2*H_ref)*C0_ref))

def main():
    siz = 2000e-9
    lam_max = .2*siz

    size = (siz, siz)
    hurst = .75
    h_rms = 3.24e-8
    res = 128
    nb_grid_pts = (res, res)

    seed = 2

    surface = Tools.RandomSurfaceGaussian(
        nb_grid_pts, size, hurst, h_rms, lambda_max=lam_max, seed=seed).get_surface()

    plot_naive(surface, lam_max)

    plot_grad_C0(surface, hurst, lam_max)

    H, C0 = plot_grad_H(surface, lam_max)
    print("H_ref = {}, C0_ref = {}".format(H, C0))
    compare_to_PyPy(surface, lam_max, H, C0)


if __name__ == "__main__":
    main()
    plt.show()



#
# Copyright 2020-2021 Lars Pastewka
#           2019-2020 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np

from ..Exceptions import NoReliableDataError
from ..HeightContainer import NonuniformLineScanInterface, UniformTopographyInterface
from ..Support import doi
from ..Support.Regression import make_grid


@doi('10.1016/j.apsadv.2021.100190')
def scale_dependent_statistical_property(self, func, n=1, scale_factor=None, distance=None, reliable=True,
                                         interpolation='linear', progress_callback=None,
                                         collocation='log', nb_points=None, nb_points_per_decade=10, threshold=4):
    """
    Compute statistical properties of a uniform topography at specific scales.
    The scale is specified either by `scale_factors` or `distance`. These
    properties are statistics of derivatives carried out at specific scales,
    as computed using the `derivative` pipeline function.

    The specific statistical properties is computed by the `func` argument.
    The output of `func` needs to be homogeneous, i.e. if an array is returned,
    this array must have the same size independent of the derivative data that
    is fed into `func`.

    Parameters
    ----------
    self : Topography or UniformLineScan
        Topogaphy or line scan.
    func : callable
        The function that computes the statistical properties:

            ``func(dx, dy=None) -> np.ndarray``

        A function taking the derivative in x-direction and optionally the
        derivative in y-direction (only for topographies, i.e. maps). The
        function needs to be able to ignore the second argument as a container
        can be a mixture of topographies and line scans. The function can
        return a scalar value or an array, but the array size must be fixed.
    n : int, optional
        Order of derivative. (Default: 1)
    scale_factor : float or np.ndarray
        Scale factor for rescaling the finite differences stencil. A scale
        factor of unity means the derivative is computed at the size of the
        individual pixel.
    distance : float or np.ndarray
        Characteristic distances at which the derivatives are computed. This
        is the overall length of the underlying stencil of lowest truncation
        order, not the effective grid spacing used by this stencil. (The
        corresponding scale factor is then given by distance / (n * px) where
        n is the order of the derivative and px the grid spacing.)
        If this is an array, then the statistical property is computed at each
        of these distances.
    reliable : bool, optional
        Only incorporate data deemed reliable. (Default: True)
    unit : str
        Unit of the distance array. All topographies are converted to this
        unit before the derivative is computed.
    interpolation : str, optional
        Interpolation method to use for computing derivatives at distances
        that do not equal an integer multiple of the grid spacing. Use
        'linear' for a local liner interpolation or 'fourier' for global
        Fourier interpolation. Note that Fourier interpolation carries large
        errors for nonperiodic topographies and should be used with care.
        (Default: 'linear')
    progress_callback : func, optional
        Function taking iteration and the total number of iterations as
        arguments for progress reporting. (Default: None)
    collocation : {'log', 'quadratic', 'linear', array_like}, optional
        Automatic computation of sampling grid if neigther `scale_factor` nor
        `distance` is specified. Specifying 'log' yields collocation points
        equally spaced on a log scale, 'quadratic' yields bins with
        similar number of data points and 'linear' yields linear bins.
        (Default: 'log')
    nb_points : int, optional
        Number of points for automatic sampling grid. Bins are automatically
        determined if set to None. (Default: None)
    nb_points_per_decade : int, optional
        Number of points per decade for log-spaced collocation points.
        (Default: None)
    threshold : int, optional
        Defines the minimal amount of data points of the probability distribution
        to calculate the statistical properties with func and returns a np.nan if
        the value is below the threshold.
        E.g. the scipy.stats.kstat function needs at least 4 data points to
        calculate the 4th cumulant function, otherwise it returns nan or inf
        (Default: 4)

    Returns
    -------
    statistical_fingerprint : np.ndarray or list of np.ndarray
        Array containing the result of `func`

    Examples
    --------
    This example yields the the scale-dependent derivative (equivalent to
    the autocorrelation function divided by the distance) in the x-direction:

    >>> distances, A = t.autocorrelation_from_profile(resampling_method=None)
    >>> s = t.scale_dependent_statistical_property(lambda x, y=None: np.var(x), distance=distances[1::20])
    >>> np.testing.assert_allclose(2 * A[1::20] / distances[1::20] ** 2, s)
    """
    if scale_factor is None and distance is None:
        lower, upper = self.bandwidth()
        distance, _ = make_grid(collocation, n * lower, upper, nb_points=nb_points,
                                nb_points_per_decade=nb_points_per_decade)
    d = self.derivative(n=n, scale_factor=scale_factor, distance=distance, interpolation=interpolation,
                        progress_callback=progress_callback)
    if distance is None:
        distance = scale_factor * np.mean(self.physical_sizes)
    if self.dim == 1:
        try:
            if scale_factor is not None:
                iter(scale_factor)
            if distance is not None:
                iter(distance)
            # check if the amount of data points fulfill threshold, otherwise set to nan
            for i in range(0, len(d)):
                if len(d[i]) < threshold:
                    d[i] = np.nan
            retvals = np.array([func(_d) for _d in d])
        except TypeError:
            # If there is a single distance, the reliability analysis is skipped
            if len(d) < threshold:
                d = np.nan
            return distance, func(d)
    else:
        try:
            if scale_factor is not None:
                iter(scale_factor)
            if distance is not None:
                iter(distance)
            # check if the amount of data points fulfill threshold, otherwise set to nan
            for i in range(0, len(d)):
                for j in range(0, len(d[i])):
                    if len(d[i][j].flatten()) < threshold:
                        d[i][j] = np.nan
            retvals = np.array([func(dx, dy) for dx, dy in zip(*d)])
        except TypeError:
            # If there is a single distance, the reliability analysis is skipped
            for i in range(0, len(d)):
                if len(d[i]) < threshold:
                    d[i] = np.nan
            return distance, func(*d)

    short_cutoff = self.short_reliability_cutoff() if reliable else None
    if short_cutoff is not None:
        mask = distance > short_cutoff * n / 2
        if mask.sum() == 0:
            raise NoReliableDataError('Dataset contains no reliable data.')
        distance = distance[mask]
        retvals = retvals[mask]

    return distance, retvals


UniformTopographyInterface.register_function(
    'scale_dependent_statistical_property', scale_dependent_statistical_property)
NonuniformLineScanInterface.register_function(
    'scale_dependent_statistical_property', scale_dependent_statistical_property)



#
# Copyright 2016, 2018-2020 Lars Pastewka
#           2019 Antoine Sanner
#           2019 Michael Röttger
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Helper functions to compute trends of surfaces
"""

import numpy as np


def tilt_from_height(topography, full_output=False):
    """
    Compute the tilt plane that if subtracted minimizes the rms height of the
    surface. The tilt plane is parameterized as:

    .. math::

        p(x, y) = h_0 + m x + n y

    The values of :math:`m`, :math:`n` and :math:`h0` are return by this
    function.

    idea as follows

    1) arr = arr_out + (ň.x + d)/ň_z
    2) arr_out.sum() = 0
    3) |ň| = 1
    => n_z = sqrt(1 - n_x^2 - n_y^2) (for 2D, but you get the idea)
       dofs = n_x, n_y, d = X

    solution X_s = arg_min ((arr - ň.x + d)^2).sum()

    Parameters
    ----------
    arr : UniformTopography
        Height information.

    Returns
    -------
    m : float
        Slope in x-direction.
    n : float
        Slope in y-direction.
    h0 : float
        Mean value.
    """
    arr = topography.heights()
    nb_dim = len(arr.shape)
    x_grids = (np.arange(arr.shape[i]) / arr.shape[i] for i in range(nb_dim))
    if nb_dim > 1:
        x_grids = np.meshgrid(*x_grids, indexing='ij')
    if np.ma.getmask(arr) is np.ma.nomask:
        columns = [x.reshape((-1, 1)) for x in x_grids]
    else:
        columns = [x[np.logical_not(arr.mask)].reshape((-1, 1))
                   for x in x_grids]
    columns.append(np.ones_like(columns[-1]))
    # linear regression model
    location_matrix = np.hstack(columns)
    offsets = np.ma.compressed(arr)
    # res = scipy.optimize.nnls(location_matrix, offsets)
    res = np.linalg.lstsq(location_matrix, offsets, rcond=None)
    coeffs = np.array(res[0])
    if full_output:
        return coeffs, location_matrix
    else:
        return coeffs


def tilt_and_curvature(arr, full_output=False):
    """
    Data in arr is interpreted as height information of a tilted and shifted
    surface.

    idea as follows

    1) arr = arr_out + (ň.x + d)/ň_z
    2) arr_out.sum() = 0
    3) |ň| = 1
    => n_z = sqrt(1 - n_x^2 - n_y^2) (for 2D, but you get the idea)
       dofs = n_x, n_y, d = X

    solution X_s = arg_min ((arr - ň.x + d)^2).sum()

    Returns:
    ---------
    coeffs [, location_matrix (if full_output)]

    coeffs ordered as follows

    {5} + {0} x + {1} y + {2} x^2 + {3} y^2 + {4} xy

    """
    arr = arr[...]
    nb_dim = len(arr.shape)
    assert nb_dim == 2
    x_grids = (np.arange(arr.shape[i]) / arr.shape[i] for i in range(nb_dim))
    # Linear terms
    x_grids = np.meshgrid(*x_grids, indexing='ij')
    # Quadratic terms
    x, y = x_grids
    x_grids += [x * x, y * y, x * y]
    if np.ma.getmask(arr) is np.ma.nomask:
        columns = [x.reshape((-1, 1)) for x in x_grids]
    else:
        columns = [x[np.logical_not(arr.mask)].reshape((-1, 1))
                   for x in x_grids]
    columns.append(np.ones_like(columns[-1]))
    # linear regression model
    location_matrix = np.hstack(columns)
    offsets = np.ma.compressed(arr)
    # res = scipy.optimize.nnls(location_matrix, offsets)
    res = np.linalg.lstsq(location_matrix, offsets, rcond=None)
    coeffs = np.array(res[0])
    if full_output:
        return coeffs, location_matrix
    else:
        return coeffs


def shift_and_tilt(topography):
    """
    returns an array of same shape and physical_sizes as arr, but shifted and
    tilted so that mean(arr) = 0 and mean(arr**2) is minimized
    """
    arr = topography.heights()
    coeffs, location_matrix = tilt_from_height(topography, full_output=True)
    coeffs = np.array(coeffs)
    offsets = arr.reshape((-1,))

    return (offsets - location_matrix @ coeffs).reshape(arr.shape)



#
# Copyright 2018-2020 Lars Pastewka
#           2019-2020 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Compare windowed parameter recoups to non-windowed
"""

import numpy as np
import scipy
import os.path

import matplotlib.pyplot as plt

import PyCo.Tools as Tools
import SurfaceTopography as Surf

counter = 0
def process(surf_name, surface):
    surfs = dict()
    surfs["Hanning "] = Tools.CharacteriseSurface(surface)
    surfs["Kaiser 8.6"] = Tools.CharacteriseSurface(surface, 'kaiser', {"beta": 8.6})
    surfs["periodic"] = Tools.CharacterisePeriodicSurface(surface)
    fig = plt.figure()
    ax=fig.add_subplot(111)
    ax.set_yscale('log')
    ax.set_xscale('log')
    colors = ('r', 'b', 'g')
    q_max = 4e8
    q_min = 2e7
    lambda_min = 2*np.pi/q_max


    names = ('Hanning ', 'Kaiser 8.6', 'periodic' )
    #names = ('periodic', )
    for color, name in zip(colors, names):
        surf = surfs[name]
        q = surf.q
        C = surf.C
        fit_sl = q>0#np.logical_and(q > q_min, q < q_max)
        global counter
        counter += 1
        print('Hello {}'.format(counter))
        #H, alpha, res = surf.estimate_hurst_alt(lambda_min=lambda_min, full_output=True, H_bracket=(0, 3))
        H, alpha = surf.estimate_hurst(lambda_min=lambda_min, full_output=True)
        title="{}: H={:.2f}, h_rms={:.2e}".format(surf_name, H, np.sqrt((surface.heights() ** 2).mean()))
        print("for '{}': Hurst = {}, C0 = {}".format(title, H, alpha))
        #print(res)
        sl = C>0
        ax.loglog(q[sl], C[sl], alpha=.1, color=color)
        mean, err, q_g = surf.grouped_stats(100)

        ax.errorbar(q_g, mean, yerr=err, color=color)
        ax.set_title(title)
        a, b = np.polyfit(np.log(q), np.log(C), 1)
        ax.plot(q[fit_sl], q[fit_sl]**(-2-2*H)*alpha, label="{}, H={:.2f}".format(name, H), color=color, lw=3)
    slice_fig = plt.figure()
    slice_ax1_4 = slice_fig.add_subplot(311)
    slice_ax1_2 = slice_fig.add_subplot(312)
    slice_ax3_4 = slice_fig.add_subplot(313)
    slices = ((slice_ax1_4, .25),
              (slice_ax1_2, .50),
              (slice_ax3_4, .75))
    for sl_ax, location in slices:
        q = surfs[names[0]].q
        q_center = location*q[0] + (1-location)*q[-1]
        slice = np.logical_and(q>.95*q_center, q< 1.05*q_center)
        Cs = [np.sqrt(surfs[name].C[slice]) for name in names]
        sl_ax.hist(Cs, bins = 50, normed = True, label = names)
        sl_ax.set_xlabel("|h(q)| for {} at q = {:.2e}".format(surf_name, q_center))

    ax.legend(loc='best')
    sl_ax.legend(loc='best')
    slice_fig.subplots_adjust(hspace=.3)
    ax.grid(True)

def plot_distro(name, surf):
    fig = plt.figure()
    ax=fig.add_subplot(111)
    ax.hist(1e9*surf.ravel(), normed=True, bins = 100, edgecolor='none', alpha=.5)
    loc, scale = scipy.stats.norm.fit(1e9*surf.ravel())
    print(scale)
    x = np.linspace(surf.min(), surf.max(), 100)*1e9
    ax.plot(x, scipy.stats.norm.pdf(x, loc, scale), label=r'Fit, $\sqrt{\sigma} = ' + "{:.2f}$ nm".format(scale))
    ax.set_xlabel("height in [nm])")
    ax.set_ylabel("probability density")
    fig.suptitle(name)
    ax.legend(loc='best')

    fig2 = plt.figure()
    ax = fig2.add_subplot(111, aspect='equal')
    ax.pcolormesh(surf)
    ax.set_xlim(0, 1024)
    ax.set_ylim(0, 1024)

def main():
    siz = 2000e-9
    size = (siz, siz)
    path = os.path.join(os.path.dirname(__file__), "SurfaceExampleUnfiltered.asc")
    surface = Surf.read_matrix(path, physical_sizes=size, factor=1e-9)
    surfs = []
    surfs.append(('Topo1', surface))
    arr, x, residual = Tools.shift_and_tilt(surface.heights(), full_output=True)
    print("ň = {[0]:.15e}, d = {}, residual = {}, mean(arr) = {}".format(
        (float(x[0]), float(x[1]), float(np.sqrt(1-x[0]**2 - x[1]**2))),
        float(x[-1]), residual, arr.mean()))
    arr = Tools.shift_and_tilt_approx(surface.heights())
    surface = Surf.Topography(arr, physical_sizes=size)
    plot_distro('Topo1_corr', surface.heights())
    surfs.append(('Topo1_corr', surface))

    hurst = .85
    res = surface.nb_grid_pts
    h_rms = 2.11e-8

    surface = Tools.RandomSurfaceGaussian(res, size, hurst, h_rms).get_surface()
    surfs.append(('Gauss periodic', surface))

    dsize = (2*siz, 2*siz)
    dres = (res[0], res[0])
    dsurface = Tools.RandomSurfaceGaussian(dres, dsize, hurst, h_rms).get_surface()
    surface = Surf.Topography(dsurface.heights()[:res[0], :res[0]], physical_sizes= size)
    surfs.append(('Gauss aperiodic', surface))

    for name, surf in surfs:#(surfs[-1],):
        process(name, surf)




if __name__ == "__main__":
    main()
    plt.show()



#
# Copyright 2021 Lars Pastewka
#           2021 Michael Röttger
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Reader for Keyence ZON files.
"""

# Thanks to @mcmalburg (https://github.com/mcmalburg) for reverse engineering the
# format. See discussion here https://github.com/gabeguss/Keyence/issues/2

import numpy as np
from numpy.lib.stride_tricks import as_strided
from struct import unpack
from zipfile import ZipFile

import defusedxml.ElementTree as ElementTree

from ..Exceptions import MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography

from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo

# The files within ZON (zip) files are named using UUIDs. Some of these
# UUIDs are fixed and contain the same information in each of these files.

# This file contains height data
HEIGHT_DATA_UUID = '4cdb0c75-5706-48cc-a9a1-adf395d609ae'

# This contains information on unit conversion
UNIT_UUID = '686613b8-27b5-4a29-8ffc-438c2780873e'

# This contains an inventory of *image* data
INVENTORY_UUID = '772e6d38-40aa-4590-85d3-b041fa243570'


def _read_array(f, dtype=np.dtype('<i4')):
    """
    Read binary array contained in a ZON archive

    Arguments
    ---------
    f : file object
        Stream to read array from
    dtype : numpy.dtype, optional
        Data type of individual element.
        (Default: 32-bit integer)
    """
    width, height, element_size, row_bytes = unpack('iiii', f.read(16))
    if element_size % dtype.itemsize != 0:
        raise ValueError(f'File report element size of {element_size} bytes, '
                         f'but requested data type requires {dtype.itemsize} bytes.')
    if row_bytes % dtype.itemsize != 0:
        raise ValueError(f'File reports {row_bytes} bytes per row, but this is not an integer multiple of the data '
                         f'type of size {dtype.itemsize} bytes.')
    raw_data = np.frombuffer(f.read(element_size * height * (row_bytes // dtype.itemsize)), dtype)

    nb_entries = element_size // dtype.itemsize
    if nb_entries == 1:
        array_data = as_strided(raw_data, shape=(width, height), strides=(dtype.itemsize, row_bytes))
    else:
        array_data = as_strided(raw_data, shape=(width, height, nb_entries),
                                strides=(dtype.itemsize, element_size, row_bytes))
    return array_data


class ZONReader(ReaderBase):
    _format = 'zon'
    _name = 'Keyence ZON'
    _description = '''
This reader open ZON files that are written by some Keyence instruments.
'''

    # Reads in the positions of all the data and metadata
    def __init__(self, file_path):
        self._file_path = file_path

        # ZON files are ZIP files with a header. The header contains a
        # thumbnail of the measurement and we are not really interested
        # in that one. Python's ZipFile automatically skips that header.

        self._channels = []
        with OpenFromAny(self._file_path, 'rb') as f:
            # ZipFile gracefully skips ZON header information before the
            # zip actually starts.
            with ZipFile(f, 'r') as z:
                # Parse unit information
                root = ElementTree.parse(z.open(UNIT_UUID)).getroot()
                meter_per_pixel = float(root.find('XYCalibration').find('MeterPerPixel').text)
                meter_per_unit = float(root.find('ZCalibration').find('MeterPerUnit').text)

                self._orig_height_scale_factor = meter_per_unit

                # Parse height data information
                # Header consists of four int32, followed by image data
                width, height, element_size = unpack('iii', z.open(HEIGHT_DATA_UUID).read(12))
                assert element_size == 4
                self._channels += [
                    ChannelInfo(self, 0, name='default', dim=2, nb_grid_pts=(width, height),
                                physical_sizes=(width * meter_per_pixel, height * meter_per_pixel),
                                height_scale_factor=self._orig_height_scale_factor, unit='m',
                                uniform=True,
                                info={'data_uuid': HEIGHT_DATA_UUID,
                                      'meter_per_pixel': meter_per_pixel,
                                      'meter_per_unit': meter_per_unit})]

    @property
    def channels(self):
        return self._channels

    def topography(self, channel_index=None, physical_sizes=None, height_scale_factor=None, unit=None, info={},
                   periodic=False, subdomain_locations=None, nb_subdomain_grid_pts=None):
        if channel_index is None:
            channel_index = self._default_channel_index

        if subdomain_locations is not None or nb_subdomain_grid_pts is not None:
            raise RuntimeError('This reader does not support MPI parallelization.')

        channel_info = self._channels[channel_index]
        physical_sizes = self._check_physical_sizes(physical_sizes, channel_info.physical_sizes)

        info.update(channel_info.info)

        if unit is not None:
            raise MetadataAlreadyFixedByFile('unit')
        unit = channel_info.unit

        with OpenFromAny(self._file_path, 'rb') as f:
            # Read image data
            with ZipFile(f, 'r') as z:
                with z.open(channel_info.info['data_uuid']) as f:
                    height_data = _read_array(f)

        topo = Topography(height_data, physical_sizes, unit=unit, info=info, periodic=periodic)

        if height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')

        return topo.scale(self._orig_height_scale_factor)



#
# Copyright 2019-2021 Lars Pastewka
#           2020-2021 Michael Röttger
#           2019 Antoine Sanner
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
In MPI Parallelized programs:

- we have to use `MPI.File.Open` instead of `open` to allow several processors
  to access the same file simultaneously
- make the file reading in 3 steps:
    - read the nb_grid_pts only (Reader.__init__)
    - make the domain decomposition according to the nb_grid_pts
    - load the relevant subdomain on each processor in Reader.topography()
"""

from ..Exceptions import FileFormatMismatch
from ..UniformLineScanAndTopography import Topography
from .Reader import ReaderBase, ChannelInfo

from NuMPI import MPI
import NuMPI
import NuMPI.IO


class NPYReader(ReaderBase):
    """
    NPY is a file format made specially for numpy arrays. They contain no extra
    metadata so we use directly the implementation from numpy and NuMPI.

    For a description of the file format, see here:
    https://docs.scipy.org/doc/numpy/reference/generated/numpy.lib.format.html
    """

    _format = 'npy'
    _name = 'Numpy arrays (NPY)'
    _description = '''
Load topography information stored as a numpy array. The numpy array format is
specified
[here](https://numpy.org/devdocs/reference/generated/numpy.lib.format.html).
The reader expects a two-dimensional array and interprets it as a map of
heights. Numpy arrays do not store units or physical sizes. These need to be
manually provided by the user.
    '''

    def __init__(self, fn, communicator=MPI.COMM_WORLD):
        """
        Open file in the NPY format.

        Parameters
        ----------
        fn : str
            Name of the file
        communicator : mpi4py MPI communicator or NuMPI stub communicator
            MPI communicator object for parallel loads.
        """
        super().__init__()

        try:
            self.mpi_file = NuMPI.IO.make_mpi_file_view(fn, communicator,
                                                        format="npy")
            self.dtype = self.mpi_file.dtype
            self._nb_grid_pts = self.mpi_file.nb_grid_pts
        except NuMPI.IO.MPIFileTypeError:
            raise FileFormatMismatch()

        # TODO: maybe implement extras specific to SurfaceTopography, like
        #  loading the units and the physical_sizes

    @property
    def channels(self):
        return [ChannelInfo(self, 0,
                            name='Default',
                            dim=len(self._nb_grid_pts),
                            uniform=True,
                            nb_grid_pts=self._nb_grid_pts)]

    def topography(self, channel_index=None, physical_sizes=None, height_scale_factor=None, unit=None, info={},
                   periodic=False, subdomain_locations=None, nb_subdomain_grid_pts=None):

        if channel_index is not None and channel_index != 0:
            raise ValueError('`channel_index` must be None or 0.')

        physical_sizes = self._check_physical_sizes(physical_sizes)
        if subdomain_locations is None and nb_subdomain_grid_pts is None:
            if self.mpi_file.comm.size > 1:
                raise ValueError("This is a parallel run, you should provide "
                                 "subdomain location and number of grid "
                                 "points")
            topography = Topography(
                heights=self.mpi_file.read(
                    subdomain_locations=subdomain_locations,
                    nb_subdomain_grid_pts=nb_subdomain_grid_pts),
                physical_sizes=physical_sizes,
                periodic=periodic,
                unit=unit,
                info=info
            )
        else:
            topography = Topography(
                heights=self.mpi_file.read(
                    subdomain_locations=subdomain_locations,
                    nb_subdomain_grid_pts=nb_subdomain_grid_pts),
                decomposition="subdomain",
                subdomain_locations=subdomain_locations,
                nb_grid_pts=self._nb_grid_pts,
                communicator=self.mpi_file.comm,
                physical_sizes=physical_sizes,
                periodic=periodic,
                unit=unit,
                info=info)

        if height_scale_factor is not None:
            topography = topography.scale(height_scale_factor)

        return topography

    channels.__doc__ = ReaderBase.channels.__doc__
    topography.__doc__ = ReaderBase.topography.__doc__


def save_npy(fn, topography):
    NuMPI.IO.save_npy(fn=fn, data=topography.heights(),
                      subdomain_locations=topography.subdomain_locations,
                      nb_grid_pts=topography.nb_subdomain_grid_pts,
                      comm=topography.communicator)



#
# Copyright 2015-2016, 2018-2021 Lars Pastewka
#           2018-2020 Antoine Sanner
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Helper functions for the generation of random fractal surfaces
"""

import numpy as np
from . import Topography, UniformLineScan


def _irfft2(karr, rarr, progress_callback=None):
    """
    Inverse 2d real-to-complex FFT with callback that allows to report status
    of the computation to user.

    Parameters
    ----------
    karr : array_like
        Fourier-space representation
    rarr : array_like
        Real-space representation
    progress_callback : function(i, n)
        Function that is called to report progress.
    """
    nrows, ncolumns = karr.shape
    for i in range(ncolumns):
        if progress_callback is not None:
            progress_callback(i, ncolumns + nrows - 1)
        karr[:, i] = np.fft.ifft(karr[:, i])
    for i in range(nrows):
        if progress_callback is not None:
            progress_callback(i + ncolumns, ncolumns + nrows - 1)
        if rarr.shape[1] % 2 == 0:
            rarr[i, :] = np.fft.irfft(karr[i, :])
        else:
            rarr[i, :] = np.fft.irfft(karr[i, :], n=rarr.shape[1])


def self_affine_prefactor(nb_grid_pts, physical_sizes, Hurst, rms_height=None,
                          rms_slope=None, short_cutoff=None, long_cutoff=None):
    r"""
    Compute prefactor :math:`C_0` for the power-spectrum density of an ideal
    self-affine topography given by

    .. math ::

        C(q) = C_0 q^{-2-2H}

    for two-dimensional topography maps and

    .. math ::

        C(q) = C_0 q^{-1-2H}

    for one-dimensional line scans. Here :math:`H` is the Hurst exponent.

    Note:
    In the 2D case:

    .. math ::

        h^2_{rms} = \frac{1}{2 \pi} \int_{0}^{\infty} q C^{iso}(q) dq

    whereas in the 1D case:

    .. math ::

        h^2_{rms} = \frac{1}{\pi} \int_{0}^{\infty} C^{1D}(q) dq

    See Equations (1) and (4) in [1].


    Parameters
    ----------
    nb_grid_pts : array_like
        Resolution of the topography map or the line scan.
    physical_sizes : array_like
        Physical physical_sizes of the topography map or the line scan.
    Hurst : float
        Hurst exponent.
    rms_height : float
        Root mean-squared height.
    rms_slope : float
        Root mean-squared slope of the topography map or the line scan.
    short_cutoff : float
        Short-wavelength cutoff.
    long_cutoff : float
        Long-wavelength cutoff.

    Returns
    -------
    prefactor : float
        Prefactor :math:`\sqrt{C_0}`

    References
    -----------
    [1]: Jacobs, Junge, Pastewka,
    Surf. Topgogr.: Metrol. Prop. 5, 013001 (2017)

    """
    nb_grid_pts = np.asarray(nb_grid_pts)
    physical_sizes = np.asarray(physical_sizes)

    if short_cutoff is not None:
        q_max = 2 * np.pi / short_cutoff
    else:
        q_max = np.pi * np.min(nb_grid_pts / physical_sizes)

    if long_cutoff is not None:
        q_min = 2 * np.pi / long_cutoff
    else:
        q_min = 2 * np.pi * np.max(1 / physical_sizes)

    area = np.prod(physical_sizes)

    if rms_height is not None:
        # Assuming no rolloff region
        fac = 2 * rms_height / np.sqrt(q_min ** (-2 * Hurst) -
                                       q_max ** (-2 * Hurst)) * np.sqrt(
            Hurst * np.pi)
    elif rms_slope is not None:
        fac = 2 * rms_slope / np.sqrt(q_max ** (2 - 2 * Hurst) -
                                      q_min ** (2 - 2 * Hurst)) * np.sqrt(
            (1 - Hurst) * np.pi)
    else:
        raise ValueError('Neither rms height nor rms slope is defined!')

    if len(nb_grid_pts) == 1:
        fac /= np.sqrt(2)

    return fac * np.prod(nb_grid_pts) / np.sqrt(area)


def fourier_synthesis(nb_grid_pts, physical_sizes,
                      hurst=None,
                      rms_height=None, rms_slope=None, c0=None,
                      psd=None,
                      short_cutoff=None, long_cutoff=None, rolloff=1.0,
                      amplitude_distribution=lambda n: np.random.normal(size=n),
                      periodic=True, rfn=None, kfn=None,
                      progress_callback=None, unit=None, info={}):
    r"""
    Create a self-affine, randomly rough surface using a Fourier filtering
    algorithm. The algorithm is described in:
    Ramisetti et al., J. Phys.: Condens. Matter 23, 215004 (2011);
    Jacobs, Junge, Pastewka, Surf. Topgogr.: Metrol. Prop. 5, 013001 (2017)

    Parameters
    ----------
    nb_grid_pts : array_like
        Resolution of the topography map.
    physical_sizes : array_like
        Physical physical_sizes of the topography map.
    hurst : float
        Hurst exponent.
    rms_height : float
        Root mean-squared height.
    rms_slope : float
        Root mean-squared slope.
    c0: float
        self affine prefactor :math:`C_0`:
        :math:`C(q) = C_0 q^{-2-2H}`
    psd: callable
        C(q)
        Function returning the Power Spectral Density as a function of the absolute value of the wavevector
    short_cutoff : float
        Short-wavelength cutoff.
    long_cutoff : float
        Long-wavelength cutoff.
    rolloff : float
        Value for the power-spectral density (PSD) below the long-wavelength
        cutoff. This multiplies the value at the cutoff, i.e. unit will give a
        PSD that is flat below the cutoff, zero will give a PSD that is
        vanishes below cutoff. (Default: 1.0)
    amplitude_distribution : function
        Function that generates the distribution of amplitudes.
        (Default: np.random.normal)
    periodic : bool
        Return a periodic surface. (Default: True)
    rfn : str
        Name of file that stores the real-space array. If specified, real-space
        array will be created as a memory mapped file. This is useful for
        creating very large topography maps. (Default: None)
    kfn : str
        Name of file that stores the Fourie-space array. If specified,
        real-space array will be created as a memory mapped file. This is
        useful for creating very large topography maps. (Default: None)
    progress_callback : function(i, n)
        Function that is called to report progress.
    unit : str, optional
        Length unit.
    info : dict, optional
        Initial info dictionary.

    Returns
    -------
    topography : UniformTopography or UniformLineScan
        The topography.
    """
    if short_cutoff is not None:
        q_max = 2 * np.pi / short_cutoff
    else:
        q_max = np.pi * np.min(np.asarray(nb_grid_pts) / np.asarray(physical_sizes))

    if long_cutoff is not None:
        q_min = 2 * np.pi / long_cutoff
    else:
        q_min = None

    if psd and (long_cutoff or hurst or rms_slope or rms_height or c0):
        raise NotImplementedError("long_cutoff, hurst, rms_slope, rms_height, c0 "
                                  "cannot be combined with prescribed PSD")

    if psd is not None:
        fac = np.prod(nb_grid_pts) / np.sqrt(np.prod(physical_sizes))
    elif c0 is None:
        fac = self_affine_prefactor(nb_grid_pts, physical_sizes, hurst, rms_height=rms_height,
                                    rms_slope=rms_slope, short_cutoff=short_cutoff,
                                    long_cutoff=long_cutoff)
    else:
        # prefactor for the fourier heights
        fac = np.sqrt(c0) * np.prod(nb_grid_pts) / np.sqrt(np.prod(physical_sizes))
        #                   ^                       ^ C(q) = c0 q^(-2-2H) = 1 / A |fh(q)|^2
        #                   |                         and h(x,y) = sum(1/A fh(q) e^(iqx)))
        #                   compensate for the np.fft normalisation

    if len(nb_grid_pts) == 2:
        nx, ny = nb_grid_pts
        sx, sy = physical_sizes
        kny = ny // 2 + 1
        kshape = (nx, kny)
    else:
        nx = 1
        ny, = nb_grid_pts
        sx = 1
        sy, = physical_sizes
        kny = ny // 2 + 1
        kshape = (kny,)

    # Create in-memory or memory-mapped arrays as storage buffers
    if rfn is None:
        rarr = np.empty(nb_grid_pts, dtype=np.float64)
    else:
        rarr = np.memmap(rfn, np.float64, 'w+', shape=nb_grid_pts)
    if kfn is None:
        karr = np.empty(kshape, dtype=np.complex128)
    else:
        karr = np.memmap(kfn, np.complex128, 'w+', shape=kshape)

    qy = 2 * np.pi * np.arange(kny) / sy
    for x in range(nx):
        if progress_callback is not None:
            progress_callback(x, nx - 1)
        if x > nx // 2:
            qx = 2 * np.pi * (nx - x) / sx
        else:
            qx = 2 * np.pi * x / sx
        q_sq = qx ** 2 + qy ** 2
        if x == 0 and not psd:
            q_sq[0] = 1.
        phase = np.exp(2 * np.pi * np.random.rand(kny) * 1j)
        ran = fac * phase * amplitude_distribution(kny)
        if psd is None:
            if len(nb_grid_pts) == 2:
                karr[x, :] = ran * q_sq ** (-(1 + hurst) / 2)
                karr[x, q_sq > q_max ** 2] = 0.
            else:
                karr[:] = ran * q_sq ** (-(0.5 + hurst) / 2)
                karr[q_sq > q_max ** 2] = 0.

        else:
            if len(nb_grid_pts) == 2:
                karr[x, :] = ran * np.sqrt(psd(np.sqrt(q_sq)))
                karr[x, q_sq > q_max ** 2] = 0.
            else:
                karr[:] = ran * np.sqrt(psd(np.sqrt(q_sq)))
                karr[q_sq > q_max ** 2] = 0.

        if q_min is not None:
            mask = q_sq < q_min ** 2
            if len(nb_grid_pts) == 2:
                karr[x, mask] = rolloff * ran[mask] * q_min ** (-(1 + hurst))
            else:
                karr[mask] = rolloff * ran[mask] * q_min ** (-(0.5 + hurst))
    if len(nb_grid_pts) == 2:
        for iy in [0, -1] if ny % 2 == 0 else [0]:
            # Enforce symmetry
            if nx % 2 == 0:
                karr[0, iy] = np.real(karr[0, iy])
                karr[nx // 2, iy] = np.real(karr[nx // 2, iy])
                karr[1:nx // 2, iy] = karr[-1:nx // 2:-1, iy].conj()
            else:
                karr[0, iy] = np.real(karr[0, iy])
                karr[1:nx // 2 + 1, iy] = karr[-1:nx // 2:-1, iy].conj()
        _irfft2(karr, rarr, progress_callback)
        return Topography(rarr, physical_sizes, periodic=periodic, unit=unit, info=info)
    else:
        karr[0] = np.real(karr[0])
        rarr[:] = np.fft.irfft(karr, n=ny)
        return UniformLineScan(rarr, sy, periodic=periodic, unit=unit, info=info)



#
# Copyright 2019-2020 Antoine Sanner
#           2018-2020 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Special topographies
"""

import numpy as np

from NuMPI import MPI
from NuMPI.Tools import Reduction

from .UniformLineScanAndTopography import Topography, UniformLineScan, \
    DecoratedUniformTopography


def make_topography_from_function(fun, physical_sizes,
                                  nb_grid_pts=None,
                                  subdomain_locations=None,
                                  nb_subdomain_grid_pts=None,
                                  centre=(0, 0),
                                  **kwargs
                                  ):
    """
    Parameters
    ----------
    fun: callable(X, Y)
    centre: tuple or "middle"
        (cx, cy)
        h(X,Y) =  fun(X-cx, Y-cy)
        default (0,0)
    physical_sizes : tuple of floats
        Physical physical_sizes of the topography map
    periodic : bool
        The topography is periodic. (Default: False)
    decomposition : str
        Specification of the data decomposition of the heights array. If
        set to 'subdomain', the heights array contains only the part of
        the full topography local to the present MPI process. If set to
        'domain', the heights array contains the global array.
        Default: 'serial', which fails for parallel runs.
    nb_grid_pts : tuple of ints
        Number of grid points for the full topography. This is only
        required if decomposition is set to 'subdomain'.
    subdomain_locations : tuple of ints
        Origin (location) of the subdomain handled by the present MPI
        process.
    nb_subdomain_grid_pts : tuple of ints
        Number of grid points within the subdomain handled by the present
        MPI process. This is only required if decomposition is set to
        'domain'.
    communicator : mpi4py communicator or NuMPI stub communicator
        The MPI communicator object.
        default value is COMM_SELF because sometimes NON-MPI readers that
        do not set the communicator value are used in MPI Programs.
        See discussion in issue #166
    info : dict
        The info dictionary containing auxiliary data. This data is never
        used by PyCo but can be used by third-party codes.

    Returns
    -------

    SurfaceTopography.Topography instance

    """
    if isinstance(centre, str) and centre == "middle":
        centre = tuple([s / 2 for s in physical_sizes])

    if nb_subdomain_grid_pts is None:
        # serial code
        nb_subdomain_grid_pts = nb_grid_pts
        nb_grid_pts = None
    topography = Topography(heights=np.zeros(nb_subdomain_grid_pts),
                            physical_sizes=physical_sizes,
                            subdomain_locations=subdomain_locations,
                            nb_grid_pts=nb_grid_pts,
                            decomposition="subdomain",
                            **kwargs,
                            )

    cx, cy = centre
    x, y = topography.positions()

    topography._heights = fun(x - cx, y - cy)

    return topography


def make_sphere(radius, nb_grid_pts, physical_sizes, centre=None,
                standoff="undefined",
                offset=0, periodic=False, kind="sphere",
                nb_subdomain_grid_pts=None, subdomain_locations=None,
                communicator=MPI.COMM_WORLD):
    r"""
    Simple sphere geometry.

    If kind="sphere" (Default)

    .. math:: h = \left\{ \begin{array}{ll} \sqrt{\text{radius}^2 - r^2} -
                  \text{radius} & \text{  for  } r < \text{radius} \\ - \text{standoff}
                   & \text{else} \end{array} \right.

    If kind="paraboloid" the sphere is approximated by a paraboloid

    .. math:: h = \frac{r^2}{2 \cdot \text{radius}}

    :math:`r^2 = x^2 + y^2`

    Parameters
    ----------
    radius : float
        self-explanatory
    nb_grid_pts : float
        self-explanatory
    physical_sizes : float
        self-explanatory
    centre : float
         specifies the coordinates (in length units, not pixels).
         by default, the sphere is centred in the topography
    kind: str
        Options are "sphere" or "paraboloid". Default is "sphere".
    standoff : float or "undefined"
         when using interaction forces with ranges of the order
         the radius, you might want to set the topography outside of
         the sphere to far away, maybe even pay the price of inf,
         if your interaction has no cutoff

         For `standoff="undefined"`, the entries will be masked in the 
         topography and topography.has_undefined_data will be true

         If `kind="paraboloid"` the paraboloid approximation is used
            and the standoff is not applied
    periodic : bool
         whether the sphere can wrap around. tricky for large spheres
    communicator : mpi4py communicator NuMPI stub communicator
         MPI communicator object.
    """  # noqa: E501, W291
    if not hasattr(nb_grid_pts, "__iter__"):
        nb_grid_pts = (nb_grid_pts,)
    dim = len(nb_grid_pts)
    if not hasattr(physical_sizes, "__iter__"):
        physical_sizes = (physical_sizes,)
    if centre is None:
        centre = np.array(physical_sizes) * .5
    if not hasattr(centre, "__iter__"):
        centre = (centre,)

    if nb_subdomain_grid_pts is None:
        nb_subdomain_grid_pts = nb_grid_pts
    if subdomain_locations is None:
        subdomain_locations = (0, 0)
    if not periodic:
        def get_r(res, size, centre, subd_loc, subd_res):
            " computes the non-periodic radii to evaluate"
            x = (subd_loc + np.arange(subd_res)) * size / res
            return x - centre
    else:
        def get_r(res, size, centre, subd_loc, subd_res):
            " computes the periodic radii to evaluate"
            x = (subd_loc + np.arange(subd_res)) * size / res
            return (x - centre + size / 2) % size - size / 2

    if dim == 1:
        r2 = get_r(nb_grid_pts[0], physical_sizes[0], centre[0],
                   subdomain_locations[0], nb_subdomain_grid_pts[0]) ** 2
    elif dim == 2:
        rx2 = (get_r(nb_grid_pts[0], physical_sizes[0], centre[0],
                     subdomain_locations[0],
                     nb_subdomain_grid_pts[0]) ** 2).reshape((-1, 1))
        ry2 = (get_r(nb_grid_pts[1], physical_sizes[1], centre[1],
                     subdomain_locations[1], nb_subdomain_grid_pts[1])) ** 2
        r2 = rx2 + ry2
    else:
        raise ValueError("Problem has to be 1- or 2-dimensional. "
                         "Yours is {}-dimensional".format(dim))

    if kind == "sphere":
        radius2 = radius ** 2  # avoid nans for small radii
        outside = r2 > radius2
        r2[outside] = radius2
        h = np.sqrt(radius2 - r2) - radius
        if standoff == "undefined":
            standoff_val = np.nan
        else:
            standoff_val = - standoff - radius
        h[outside] = standoff_val
    elif kind == "paraboloid":
        h = - r2 / (2 * radius)
    else:
        raise (ValueError("Wrong value given for parameter kind {}. "
                          "Should be 'sphere' or 'paraboloid'".format(kind)))

    if dim == 1:
        ret_top = UniformLineScan(h + offset, physical_sizes)
    else:
        ret_top = Topography(h + offset, physical_sizes,
                             decomposition='subdomain',
                             nb_grid_pts=nb_grid_pts,
                             subdomain_locations=subdomain_locations,
                             communicator=communicator)

    if standoff == "undefined":
        return ret_top
    else:
        return ret_top.fill_undefined_data(standoff_val)


class PlasticTopography(DecoratedUniformTopography):
    """ SurfaceTopography with an additional plastic deformation field.
    """
    name = 'plastic_topography'

    def __init__(self, topography, hardness, plastic_displ=None):
        """
        Keyword Arguments:
        topography -- topography profile
        hardness -- penetration hardness
        plastic_displ -- initial plastic displacements
        """
        super().__init__(topography)
        self.hardness = hardness
        if plastic_displ is None:
            plastic_displ = np.zeros(self.nb_subdomain_grid_pts)
        self.plastic_displ = plastic_displ

    def __getstate__(self):
        """ is called and the returned object is pickled as the contents for
            the instance
        """
        state = super().__getstate__(), self.hardness, self.plastic_displ
        return state

    def __setstate__(self, state):
        """ Upon unpickling, it is called with the unpickled state
        Keyword Arguments:
        state -- result of __getstate__
        """
        superstate, self.hardness, self.plastic_displ = state
        super().__setstate__(superstate)

    @property
    def hardness(self):
        return self._hardness

    @hardness.setter
    def hardness(self, hardness):
        if hardness < 0:
            raise ValueError('Hardness must be positive.')
        self._hardness = hardness

    @property
    def plastic_displ(self):
        return self.__h_pl

    @plastic_displ.setter
    def plastic_displ(self, plastic_displ):
        if plastic_displ.shape != self.nb_subdomain_grid_pts:
            raise ValueError(
                'Resolution of profile and plastic displacement must match.')
        self.__h_pl = plastic_displ

    def undeformed_profile(self):
        """ Returns the undeformed profile of the topography.
        """
        return self.parent_topography.heights()

    def heights(self):
        """ Computes the combined profile.
        """
        return self.undeformed_profile() + self.plastic_displ

    @property
    def plastic_area(self):
        pnp = Reduction(self._communicator)
        return pnp.sum(np.count_nonzero(self.__h_pl)) * self.area_per_pt



#
# Copyright 2020 Lars Pastewka
#           2020 Antoine Sanner
#           2015 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
objective function and jacobians for hurst estimation
"""
from sympy import Symbol, pprint
import sympy


q = Symbol('q', positive=True)
denom = Symbol('denom', positive=True)
C = Symbol('C', positive=True)
C0 = Symbol('C_0', positive=True)
H = Symbol('H', positive=True)
cor = Symbol('K', positive=True)


obj = (C-C0*q**(-2-2*H))**2/denom
pprint(obj)

jacob = [sympy.diff(obj, var) for var in (H, C0)]
print("jacobian:")
pprint(jacob)

sol_C0 = sympy.solve(jacob[1], C0)
pprint(sol_C0)

fprime_h = sympy.diff(jacob[0], H)

pprint(sympy.simplify(fprime_h))


pprint(sympy.diff(q**(-2-2*H), H))



#
# Copyright 2022 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

#
# Reference information and implementations:
# https://sourceforge.net/p/gwyddion/code/HEAD/tree/trunk/gwyddion/modules/file/alicona.c
#

import numpy as np

from .binary import decode
from .common import OpenFromAny
from .Reader import ReaderBase, ChannelInfo
from ..Exceptions import CorruptFile, FileFormatMismatch, MetadataAlreadyFixedByFile
from ..UniformLineScanAndTopography import Topography


class AL3DReader(ReaderBase):
    _format = 'al3d'
    _name = 'Alicona Imaging AL3D'
    _description = '''
AL3D format of Alicona Imaging.
'''

    _MAGIC = b'AliconaImaging\x00\r\n'

    _tag_structure = [
        ('key', '20s'),
        ('value', '30s'),
        ('crlf', '2s'),
    ]

    _header_structure = [
        ('nb_tags', 'I'),
        ('nb_grid_pts_x', 'I'),
        ('nb_grid_pts_y', 'I'),
        ('nb_planes', 'I'),
        ('grid_spacing_x', 'd'),
        ('grid_spacing_y', 'd'),
        ('icon_offset', 'I'),
        ('texture_offset', 'I'),
        ('depth_offset', 'I'),
    ]

    # Reads in the positions of all the data and metadata
    def __init__(self, file_path):
        self.file_path = file_path
        with OpenFromAny(file_path, 'rb') as f:
            # Detect file version
            magic = f.read(len(self._MAGIC))

            # Check AL3D file magic
            if magic != self._MAGIC:
                # This is not an AL3D file
                raise FileFormatMismatch

            version = int(self._read_tag(f, 'Version'))
            nb_tags = int(self._read_tag(f, 'TagCount'))

            self._header = {'Version': version, 'TagCount': nb_tags}
            for i in range(nb_tags):
                tag = decode(f, self._tag_structure, '<')
                if tag['crlf'] != '\r\n':
                    raise CorruptFile('CRLF tag terminator missing.')
                self._header[tag['key']] = tag['value']

            nx, ny = int(self._header['Cols']), int(self._header['Rows'])
            self._nb_grid_pts = (nx, ny)
            self._physical_sizes = (nx * float(self._header['PixelSizeXMeter']),
                                    ny * float(self._header['PixelSizeYMeter']))

        self._unit = 'm'

    def _read_tag(self, f, key):
        tag = decode(f, self._tag_structure, '<')
        if tag['key'] != key:
            raise CorruptFile('Expected tag key {}, found {}.'.format(key, tag['key']))
        if tag['crlf'] != '\r\n':
            raise CorruptFile('CRLF tag terminator missing.')
        return tag['value']

    def read_height_data(self, f):
        f.seek(int(self._header['DepthImageOffset']))
        invalid_pixel_value = float(self._header['InvalidPixelValue'])
        dtype = np.single
        buffer = f.read(np.prod(self._nb_grid_pts) * np.dtype(dtype).itemsize)
        data = np.frombuffer(buffer, dtype=dtype).reshape(self._nb_grid_pts)
        return np.ma.masked_array(data, mask=data == invalid_pixel_value)

    @property
    def channels(self):
        return [ChannelInfo(self,
                            0,  # channel index
                            name='Default',
                            dim=2,
                            nb_grid_pts=self._nb_grid_pts,
                            physical_sizes=self._physical_sizes,
                            uniform=True,
                            unit={'raw_metadata': self._header})]

    def topography(self, channel_index=None, physical_sizes=None,
                   height_scale_factor=None, unit=None, info={},
                   periodic=None, subdomain_locations=None,
                   nb_subdomain_grid_pts=None):
        if subdomain_locations is not None or \
                nb_subdomain_grid_pts is not None:
            raise RuntimeError('This reader does not support MPI parallelization.')

        if channel_index is None:
            channel_index = self._default_channel_index

        if channel_index != self._default_channel_index:
            raise RuntimeError(f'There is only a single channel. Channel index must be {self._default_channel_index}.')

        if physical_sizes is not None:
            raise MetadataAlreadyFixedByFile('physical_sizes')

        if height_scale_factor is not None:
            raise MetadataAlreadyFixedByFile('height_scale_factor')

        if unit is not None:
            raise MetadataAlreadyFixedByFile('unit')

        with OpenFromAny(self.file_path, 'rb') as f:
            height_data = self.read_height_data(f)

        _info = info.copy()
        _info['raw_metadata'] = self._header

        topo = Topography(height_data,
                          self._physical_sizes,
                          unit=self._unit,
                          periodic=False if periodic is None else periodic,
                          info=_info)
        return topo



#
# Copyright 2018-2020 Lars Pastewka
#           2019 Antoine Sanner
#           2019 Michael Röttger
#           2015-2016 Till Junge
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Bin for small common helper function and classes for containers.
"""

import SurfaceTopography.Support.UnitConversion as UnitConversion
from .SurfaceContainer import SurfaceContainer


def bandwidth(self, unit):
    """
    Computes lower and upper bound of bandwidth, i.e. of the wavelengths or
    length scales occurring on any topography within the container. The lower
    end of the bandwidth is given by the mean of the spacing of the individual
    points on the line scan. The upper bound is given by the overall length of
    the line scan.

    Parameters
    ----------
    self : SurfaceContainer
        Container object containing the underlying data sets.
    unit : str
        Unit for reporting the bandwidths.

    Returns
    -------
    lower_bound : float
        Lower bound of the bandwidth.
    upper_bound : float
        Upper bound of the bandwidth.
    """
    global_lower = global_upper = None
    for topography in self:
        current_lower, current_upper = topography.bandwidth()
        fac = UnitConversion.get_unit_conversion_factor(topography.unit, unit)
        global_lower = current_lower * fac if global_lower is None else min(global_lower, current_lower * fac)
        global_upper = current_upper * fac if global_upper is None else max(global_upper, current_upper * fac)

    return global_lower, global_upper


def suggest_length_unit(self, scale):
    """
    Compute a suggestion for a unit to diplay container-wider information.
    The unit is chose to minimize number of digits to the left and right of
    the decimal point.

    Parameters
    ----------
    scale : str
        'linear': displaying data on a linear axis
        'log' displaying data on a log-space axis

    Returns
    -------
    unit : str
        Suggestion for the length unit
    """
    return UnitConversion.suggest_length_unit(scale, *bandwidth(self, unit='m'))


# Register analysis functions from this module
SurfaceContainer.register_function('bandwidth', bandwidth)
SurfaceContainer.register_function('suggest_length_unit', suggest_length_unit)
